import 'dart:async';

import 'package:arango_driver/arango_driver.dart';
import 'package:arangodb_repository/src/not_found.dart';
import 'package:arangodb_repository/src/unautorized.dart';
import 'package:nosql_repository/nosql_repository.dart';
import 'package:tuple/tuple.dart';

import 'expression_rendering/context.dart';
import 'expression_rendering/expression_extension.dart';

class ArangoDbRepository<TEntity> implements Repository<TEntity> {
  final ArangoDBClient db;
  final String collectionName;

  ArangoDbRepository(this.db, this.collectionName);

  @override
  Future<Map<String, dynamic>> create(
    Map<String, dynamic> map,
    DbPrincipal principal, [
    String permission = '',
  ]) async {
    map['meta'] = {
      'tenantKey': principal.tenantKey,
      'revisions': [
        _createRevision(principal),
      ]
    };
    if (map.containsKey('_key') && (map['_key'] ?? '') == '') {
      map.remove('_key');
    }
    if (permission != '') {
      _authorize(map, principal, permission);
    }
    var resultMap = await db.createDocument(collectionName, map);

    var key = _handleDataResult(resultMap);
    map = (await db.getDocumentByKey(collectionName, key)).document;
    return map;
  }

  @override
  Future<Map<String, dynamic>> update(
    Map<String, dynamic> map,
    DbPrincipal principal, [
    String permission = '',
  ]) async {
    var existing = await _get(
      map['_key'],
      principal,
    );

    if (permission != '') {
      _authorize(existing, principal, permission, 'update');
    }

    var existingMeta = existing['meta'];
    existingMeta['revisions'].add(_createRevision(principal));

    map['meta'] = existingMeta;

    var resultMap = await db.updateDocument(collectionName, map['_key'], map);
    var key = _handleDataResult(resultMap);
    map = (await db.getDocumentByKey(collectionName, key)).document;

    return map;
  }

  @override
  Future delete(
    String key,
    DbPrincipal principal, [
    String permission = '',
  ]) async {
    var map = await _get(key, principal);
    if (permission != '') {
      _authorize(map, principal, permission, 'delete');
    }
    var result = await db.removeDocument(collectionName, key);
    _handleDataResult(result);
  }

  @override
  Future<Map<String, dynamic>> get(
    String key,
    DbPrincipal principal, [
    String permission = '',
  ]) async {
    final map = await _get(key, principal);
    if (permission != '') {
      _authorize(map, principal, permission, 'read');
    }

    return map;
  }

  @override
  Stream<Map<String, dynamic>> getAll(
    DbPrincipal principal, [
    String permission = '',
  ]) {
    var ret = _runQuery('', principal);
    return ret;
  }

  @override
  Stream<Map<String, dynamic>> search(
    SearchCriteria criteria,
    DbPrincipal principal, {
    String permission = '',
  }) {
    var filters = _createFilters(criteria);
    var filterQuery = filters.item1;
    var parameters = filters.item2;

    var page =
        _createPage(filterQuery, criteria, parameters, principal, permission);

    return page;
  }

  @override
  Future<int> count(
    SearchCriteria criteria,
    DbPrincipal principal, {
    String permission = '',
  }) async {
    var filters = _createFilters(criteria);
    var baseQuery = filters.item1;
    var parameters = filters.item2;
    var count = await _getCount(
      baseQuery,
      parameters,
      principal,
      permission: permission,
    );

    return count;
  }

  @override
  Future<SearchResult> searchWithCount(
    SearchCriteria criteria,
    DbPrincipal principal, {
    String permission = '',
  }) async {
    var filters = _createFilters(criteria);
    var filterQuery = filters.item1;
    var parameters = filters.item2;
    var count = await (_getCount(
      filterQuery,
      {...parameters},
      principal,
      permission: permission,
    ));

    var page = _createPage(
      filterQuery,
      criteria,
      parameters,
      principal,
      permission,
    );

    var ret = SearchResult(page, count);
    return ret;
  }

  Stream<Map<String, dynamic>> _runQuery(
    String query,
    DbPrincipal principal, {
    Map<String, dynamic> parameters = const {},
    String permission = '',
    String collectionAlias = '',
  }) {
    var q = _createQuery(
      principal,
      permission: permission,
      collectionAlias: collectionAlias,
    );

    if (query != '') {
      q = q.addLine(query);
    }

    for (var entry in parameters.entries) {
      q = q.addBindVar(entry.key, entry.value);
    }
    var stream = q.runAndReturnStream();

    return stream;
  }

  QueryWithClient _createQuery(
    DbPrincipal principal, {
    String permission = '',
    String collectionAlias = '',
  }) {
    var q = db.newQuery().addBindVar('tenantKey', principal.tenantKey);
    if (permission != '' && principal.hasPermission(permission)) {
      q = q.addLine('''let $collectionName = (
                    for c in $collectionName filter c.meta.tenantKey == @tenantKey return c
                  )''').addBindVar('tenantKey', principal.tenantKey);
    } else {
      q = q
          .addLine('''let $collectionName = (
                    for c in $collectionName 
                    filter c.meta.tenantKey == @tenantKey 
                          && length(c.meta.shares[* FILTER CURRENT.userKey == @userKey && \'read\' IN CURRENT.actions ]) > 0 
                    return c
                  )''')
          .addBindVar('tenantKey', principal.tenantKey)
          .addBindVar('userKey', principal.userKey);
    }
    if (collectionAlias != '') {
      q = q.addLine(
          'let $collectionAlias = (for c in $collectionName return c)');
    }

    return q;
  }

  Map<String, dynamic> _createRevision(DbPrincipal principal) {
    return {
      'userKey': principal.userKey,
      'revisionDate': DateTime.now().toUtc().toIso8601String()
    };
  }

  Future<Map<String, dynamic>> _get(String key, DbPrincipal principal) async {
    final response = (await db.getDocumentByKey(collectionName, key));
    if (response.result.error) {
      switch (response.result.errorNum) {
        case 1202:
          throw NotFound();
        default:
          throw Error();
      }
    }

    final map = response.document;

    if (!_isValidateTenant(principal, map)) {
      throw Unauthorized();
    }

    return map;
  }

  bool _isValidateTenant(DbPrincipal principal, Map<String, dynamic> map) {
    final entityTenantKey = (map['meta'] ?? {})['tenantKey'] as String?;

    final isValid =
        entityTenantKey != null && entityTenantKey == principal.tenantKey;
    return isValid;
  }

  void _authorize(
    Map<String, dynamic> entity,
    DbPrincipal principal,
    String permission, [
    String? sharePermission,
  ]) {
    if (principal.hasPermission(permission)) {
      return;
    }

    if ((sharePermission ?? '') != '') {
      var shares = ((entity['meta'] ?? {})['shares'] ?? []) as List;

      if (shares.any((s) =>
          s['userKey'] == principal.userKey &&
          (s['actions'] ?? []).contains(sharePermission))) {
        return;
      }
    }
    throw Unauthorized();
  }

  String _handleDataResult(OperationResult operationResult) {
    if (operationResult.result.error) {
      throw DbException(
        message: operationResult.result.errorMessage,
        code: operationResult.result.code.toString(),
        number: operationResult.result.errorNum.toString(),
      );
    }
    return operationResult.identifier.key;
  }

  Tuple2<String, Map<String, dynamic>> _createFilters(SearchCriteria criteria) {
    var context = Context();
    var buffer = StringBuffer();
    for (var condition in criteria.searchConditions) {
      var expression = condition.render(context);
      buffer.writeln('FILTER $expression');
    }
    var ret = Tuple2(buffer.toString(), context.parameters);
    return ret;
  }

  Stream<Map<String, dynamic>> _createPage(
    String filterQuery,
    SearchCriteria criteria,
    Map<String, dynamic> parameters,
    DbPrincipal principal,
    String permission,
  ) {
    final queryBuffer = StringBuffer();

    queryBuffer.writeln('FOR entity in entities ');
    queryBuffer.writeln(filterQuery);

    var sort = _createSort(criteria);
    queryBuffer.writeln(sort);

    if (criteria.skip != null || criteria.take != null) {
      queryBuffer.writeln('LIMIT @skip, @take');
      parameters.addAll({'skip': criteria.skip, 'take': criteria.take});
    }

    var retrn = _createReturn(criteria);
    queryBuffer.writeln(retrn);

    var query = queryBuffer.toString();

    var page = _runQuery(query, principal,
        parameters: parameters,
        permission: permission,
        collectionAlias: 'entities');
    return page;
  }

  String _createSort(SearchCriteria criteria) {
    var sortBuffer = StringBuffer();
    for (var f in criteria.orderByFields) {
      if (sortBuffer.isEmpty) {
        sortBuffer.write(sortBuffer.isEmpty ? 'sort' : ',');
      }
      sortBuffer.write(' entity.${_sanitizePath(f.path)}');
    }

    return sortBuffer.toString();
  }

  String _createReturn(SearchCriteria criteria) {
    if ((criteria.returnFields).isEmpty) {
      return 'RETURN entity';
    }

    var sb = StringBuffer();

    for (var rf in criteria.returnFields) {
      if (sb.isEmpty) {
        sb.write('RETURN {');
      } else {
        sb.write(', ');
      }

      var alias = rf.alias ?? rf.path!.replaceAll('.', '_');
      sb.write(' $alias : entity.${rf.path} ');
    }
    sb.write('}');

    var ret = sb.toString();
    return ret;
  }

  Future<int> _getCount(
    String query,
    Map<String, dynamic> parameters,
    DbPrincipal principal, {
    String permission = '',
  }) async {
    var countQuery = 'RETURN { cnt: length($query) }';
    var count = (await _runQuery(countQuery, principal,
            parameters: parameters,
            permission: permission,
            collectionAlias: 'entities')
        .first)['cnt'] as int;

    return count;
  }

  String _sanitizePath(String path) {
    var sanitized = path.replaceAll(RegExp(r'[^a-zA-Z0-9\.\_]'), '');
    return sanitized;
  }
}
